/**
 * Sound Pad - keyboard on palm
 * Copyright (C) 2000 by matobaa
 * @version 0.12 2000/12/15
 * @version 0.11 2000/12/10
 * @version 0.1 2000/12/08
 * @author matobaa	mailto:matobaa@mail.dddd.ne.jp
 * @see http://www2u.biglobe.ne.jp/~matobaa/palm/palmware.html
 *
 * sustain concept and UI by P-Key by Koichiro Nakagawa
 * http://hp.vector.co.jp/authors/VA017414/
 */

#include <Pilot.h>
#include <SysEvtMgr.h>
#include <KeyMgr.h>
#include "resource.h"

// From SonyChars.h
#define vchrJogUp		   0x1700
#define vchrJogDown		   0x1701

#define ourMinVersion 3
#define CenterC 60 /* by midi# */
#define mute 3 /* ticks */

#define TONE_KOEF 1.05946309435929530984 /* log2(1/12) */

/* Internal Structures */
typedef struct {} MyPreferenceType;
typedef struct {} MyAppInfoType;
enum { C = 0, Cis, D, Dis, E, F, Fis, G, Gis, A, Ais, H } ;
typedef enum { FreePlayMode, RecordMode, PlaybackMode } Mode;
typedef struct {
	short note;
	ULong startTime;
	ULong duration;
} Note;

#define Off 128

/* Internal Variables */
static SndCommandType onCmd = {sndCmdNoteOn, 60,2200,127};
static SndCommandType offCmd = {sndCmdQuiet, 0,0,0};
static MyPreferenceType prefs;
static MyAppInfoType appInfo;
static UInt offset = CenterC;
static Mode mode = RecordMode;
static VoidHand recordHandle;
static Note* record;
static UInt cursor;
static ULong startTime;
static int currentNote = Off;
static UInt recordedLength;
static ULong timeToNewNote;
static UInt maxRecordLength = 60;
static Boolean RECORDABLE;

/* Internal Constants */
static const unsigned long appFileCreator = 'SndP';
static const unsigned short appVersionNum = 0x01;
static const unsigned short appPrefID = 0x00;
static const unsigned short appPrefVersionNum = 0x01;

/* prototypes */
static Err RomVersionCompatible(DWord , Word );
static VoidPtr GetObjectPtr(Word);
static int getNote(UInt,UInt);
static void rewind(void);
static void DrawKeyboard(void);
static void startRecording(void);
static void startPlayback(void);
static void playback(void);

/* functions */

/** Check the rom version */
static Err RomVersionCompatible(DWord requiredVersion, Word launchFlags) {
	DWord romVersion;
	// See if we're on in minimum required version of the ROM or later.
	FtrGet(sysFtrCreator, sysFtrNumROMVersion, &romVersion);
	if (romVersion < requiredVersion) {
		if ((launchFlags & (sysAppLaunchFlagNewGlobals | sysAppLaunchFlagUIApp)) ==
				(sysAppLaunchFlagNewGlobals | sysAppLaunchFlagUIApp)) {
			FrmAlert (SystemIncompatible);
			// Pilot 1.0 will continuously relaunch this app unless we switch to
			// another safe one.
			if (romVersion < sysMakeROMVersion(3,0,0,sysROMStageRelease,0)) {
				AppLaunchWithCommand(sysFileCDefaultApp, sysAppLaunchCmdNormalLaunch, NULL);
			}
		}
		return sysErrRomIncompatible;
	}
	return 0;
}

/** Get the UI Oeject's Ptr specified by ID */
static VoidPtr GetObjectPtr(Word objectID) {
	FormPtr frmP;
	frmP = FrmGetActiveForm();
	return (FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, objectID)));
}

#define KBPosX 3
#define KBPosY 19
#define IvoryWidth 22
#define EbonyWidth 15
#define EbonyHeight 42
#define IvoryHeight 76
#define KBWidth (IvoryWidth * 7)
#define KBHeight IvoryHeight

/* local inline function */
#define returnIfContains(x, y, BitmapID) \
			FrmGetObjectBounds(frmP, FrmGetObjectIndex(frmP, (BitmapID)), &rect);\
			if (RctPtInRectangle((x),(y),&rect)) return (BitmapID)
#define included(x,y,w,h,tx,ty) ((x) < (tx) && (tx) < (x+w) && (y) < (ty) && (ty) < (y+h))

/** get key ID by tapped location */
static int getNote(UInt x,UInt y) {
	RectangleType rect;
	int roundedOffset;
	FormPtr frmP;
	UInt objIndex;
	frmP = FrmGetActiveForm();
	returnIfContains(x,y,PlaybackBitmap);
	returnIfContains(x,y,RewindBitmap);
	// returnIfContains(x,y,StepbackBitmap);
	// returnIfContains(x,y,StepfwdBitmap);
	// returnIfContains(x,y,LogoBitmap);
	// else
	roundedOffset = offset - offset % 12;
	x += ((offset % 12) + 1 & ~1) * IvoryWidth / 2; // re-align x by offset
	if(x > KBPosX + KBWidth) { x -= KBWidth ; roundedOffset += 12;} // fold by 3 to 156
	if(x < KBPosX) { x += KBWidth ; roundedOffset -= 12;}
	return included( 16,KBPosY,EbonyWidth-1,EbonyHeight,x,y) ? Cis + roundedOffset :
		   included( 42,KBPosY,EbonyWidth-1,EbonyHeight,x,y) ? Dis + roundedOffset :
		// included( ??.KBPosY,EbonyWidth-1,EbonyHeight,x,y) ? Eis + roundedOffset :
		   included( 81,KBPosY,EbonyWidth-1,EbonyHeight,x,y) ? Fis + roundedOffset :
		   included(106,KBPosY,EbonyWidth-1,EbonyHeight,x,y) ? Gis + roundedOffset :
		   included(131,KBPosY,EbonyWidth-1,EbonyHeight,x,y) ? Ais + roundedOffset :
		   included(  3,KBPosY,IvoryWidth,IvoryHeight,x,y) ? C + roundedOffset :
		   included( 25,KBPosY,IvoryWidth,IvoryHeight,x,y) ? D + roundedOffset :
		   included( 47,KBPosY,IvoryWidth,IvoryHeight,x,y) ? E + roundedOffset :
		   included( 69,KBPosY,IvoryWidth,IvoryHeight,x,y) ? F + roundedOffset :
		   included( 91,KBPosY,IvoryWidth,IvoryHeight,x,y) ? G + roundedOffset :
		   included(113,KBPosY,IvoryWidth,IvoryHeight,x,y) ? A + roundedOffset :
		   included(135,KBPosY,IvoryWidth,IvoryHeight,x,y) ? H + roundedOffset : Off ;
}

#undef included
#undef returnIfContains

/* get Octave shift offset */
static unsigned int TempOffset() {
	int currentState = KeyCurrentState();
	return ((currentState & keyBitHard1) != 0) * -24
		 + ((currentState & keyBitHard2) != 0) * -12
		 + ((currentState & keyBitHard3) != 0) *  12
		 + ((currentState & keyBitHard4) != 0) *  24;
}

/* local inline functions */
/*
#define noteContinue() SndDoCmd(NULL, (currentNote != Off ) ? &onCmd : &offCmd, -1);
#define noteOn() onCmd.param1 = note + TempOffset(); SndDoCmd(NULL, &onCmd, -1); currentNote = note;
#define noteOff() SndDoCmd(NULL, &offCmd, -1); currentNote = Off;
*/

/** start note */
static void noteOn(short note) {
	onCmd.param1 = note + TempOffset();
	SndDoCmd(NULL, &onCmd, -1);
	if(RECORDABLE && mode == RecordMode && currentNote != note && cursor < maxRecordLength) {
		if(startTime == 0) startTime = TimGetTicks();
		record[cursor].note = onCmd.param1;
		record[cursor].startTime = TimGetTicks() - startTime;
		cursor++;
	}
	currentNote = note;
}

/** stop note */
static void noteOff() {
	SndDoCmd(NULL, &offCmd, -1);
	currentNote = Off;
	if(RECORDABLE && mode == RecordMode && cursor < maxRecordLength && startTime != 0) {
		record[cursor].note = Off;
		record[cursor].startTime = TimGetTicks() - startTime;
		cursor++;
	}
}

/** continue current note */
static void noteContinue() {
	if(currentNote != Off)
		SndDoCmd(NULL, &onCmd, -1);
}


/** Handle events at main form */
static Boolean MainFormHandleEvent(EventPtr eventP) {
	FormPtr frmP;
	int note = getNote(eventP->screenX, eventP->screenY);
	Boolean handled = false;
	switch (eventP->eType)	{
	case frmOpenEvent:
		frmP = FrmGetActiveForm();
		FrmDrawForm ( frmP);
		KeySetMask(~(keyBitHard1 | keyBitHard2 | keyBitHard3 | keyBitHard4));
		DrawKeyboard();
		startRecording();
		handled = true;
		break;
	case frmCloseEvent:
		KeySetMask( keyBitsAll );
		handled = true;
		break;
	case keyDownEvent:
		if(eventP->data.keyDown.chr == chrPageUp ||
		   eventP->data.keyDown.chr == vchrJogUp ) {
			if(offset <= 127 - 12 - 2) {
				offset += 2;
				if(offset % 12 == 6) offset--;
				if(offset % 12 == 1) offset--;
				DrawKeyboard();
			}
		} else
		if(eventP->data.keyDown.chr == chrPageDown ||
		   eventP->data.keyDown.chr == vchrJogDown ) {
			if(offset >= 2) {
				offset -= 2;
				if(offset % 12 == 3) offset++;
				if(offset % 12 == 10) offset++;
				DrawKeyboard();
			}
		}
//	  |1|3|-|6|8|A|
//	 |0|2|4|5|7|9|B|
		handled = true;
		if (eventP->penDown &&
				(note = getNote(eventP->screenX, eventP->screenY)) > Off ) {
			noteOn(note); // now, we have new offset value: recalculate note.
		} else {
			noteContinue();
		}
		break;
	case penDownEvent:
		if (note == Off) { // tapped on outside
			noteOff();
		} else if (note == PlaybackBitmap && RECORDABLE) {
			if (mode == RecordMode) startPlayback();
			else startRecording();
		} else if (note == RewindBitmap) {
			rewind();
		} else {
			if ( note == currentNote ) { // tapped on same key
				noteOff(); // quite mute
				SysTaskDelay(mute);
			}
			noteOn(note); // play a note
		}
		handled = true;
		break;
	case penMoveEvent:
		if ( note < Off && note != currentNote) { // moved to other key
			noteOn(note);
		} else {
			noteContinue();
		}
		handled = true;
		break;
	case penUpEvent:
		if ( note == currentNote) { // if release on same key
			noteOff(); // stop note
		} else { // if release on outside
			noteContinue(); // sustain
		}
		handled = true;
		break;
	case nilEvent:
		if( RECORDABLE && mode == PlaybackMode
		     && TimGetTicks() >= timeToNewNote) {
			playback();
		} else {
			noteContinue();
		}
		handled = true;
		break;
	default:
		noteContinue();
		break;
	}
	return handled;
}

/** reset cursor to beginning */
static void rewind() {
	noteOff();
	cursor = 0;
	startTime = (mode == RecordMode   ) ? 0 :
				(mode == PlaybackMode ) ? TimGetTicks() : 0;
}

/** initiate for recording */
static void startRecording() {
	WinDrawChars("RECORDING ",10,0,140);
	mode = RecordMode;
	rewind();
}

/** initiate for playback */
static void startPlayback() {
	if(!RECORDABLE) return;
	if(mode == RecordMode && cursor != 0) { // if in recording , finish it.
		if(currentNote != Off) noteOff();
		recordedLength = cursor;
	}
	WinDrawChars("PLAYBACK  ",10,0,140);
	mode = PlaybackMode;
	rewind();
	playback();
}

/** playback recorded note */
static void playback() {
	short note;
	if(!RECORDABLE || mode != PlaybackMode) return;
	if(cursor < recordedLength) {
		note = record[cursor++].note;
		if(note != Off) noteOn(note); else noteOff();
		timeToNewNote = startTime + record[cursor].startTime;
	} // NO else. I Know this is O.K. detecting irregal situation
	if(cursor >= recordedLength) {
		// playback done. return to Record Mode.
		startRecording();
	}
}

/** Draw ebony keys in specified location */
static void WinDrawEbony(int x){
	RectangleType rect;
	while(x > KBPosX + KBWidth ) x -= KBWidth;
	rect.topLeft.x = x;  rect.extent.x = EbonyWidth;
	rect.topLeft.y = KBPosY; rect.extent.y = EbonyHeight ;
	if(x > KBPosX + KBWidth - EbonyWidth ) { // if right most key,
		// draw additional key in leftside
		rect.topLeft.x = 0;
		rect.extent.x = EbonyWidth - (KBWidth - x);
		WinDrawRectangle(&rect,1);
		// and restore location for original key
		rect.topLeft.x = x;
		rect.extent.x = (x < 160 - EbonyWidth) ? EbonyWidth : 160 - x;
	}
	WinDrawRectangle(&rect,1);
	WinEraseLine( x+2, 24, x+2, 58); // reflect
}

/** draw whole keyboard */
static void DrawKeyboard() {
	int i;
	char buffer;
	RectangleType rect;
	int roundedOffset = ((offset % 12) + 1 & ~1 )/ 2;
	// erase
	rect.topLeft.x = 0; rect.topLeft.y = KBPosY;
	rect.extent.x = 160; rect.extent.y = IvoryHeight;
	WinEraseRectangle(&rect,0);
	// Ebony
	WinDrawEbony( 16 + KBWidth - roundedOffset * IvoryWidth); // Cis
	WinDrawEbony( 42 + KBWidth - roundedOffset * IvoryWidth); // Dis
	WinDrawEbony( 81 + KBWidth - roundedOffset * IvoryWidth); // Fis
	WinDrawEbony(106 + KBWidth - roundedOffset * IvoryWidth); // Gis
	WinDrawEbony(131 + KBWidth - roundedOffset * IvoryWidth); // Ais
	// Ivory
	WinDrawLine(0,KBPosY - 1,159,KBPosY - 1);
	WinDrawLine(0,KBPosY + IvoryHeight + 1,159,KBPosY + IvoryHeight + 1);
	for(i = KBPosX ; i <= KBPosX + KBWidth; i += IvoryWidth) {
		WinDrawLine( i, KBPosY, i, KBPosY + IvoryHeight);
	}
	// C Mark
	rect.topLeft.x = (roundedOffset == 0) ? 12
					: 12 + KBWidth - roundedOffset * IvoryWidth;
	rect.topLeft.y = 85;
	rect.extent.x  = 4; rect.extent.y = 4;
	// below lines were modified because of fumiaki's suggestion. THANX!
	// WinDrawRectangle(&rect,1);
	buffer = '0' + ((offset-1) / 12); // little tricky
	WinDrawChars(&buffer,1,rect.topLeft.x, rect.topLeft.y);
}

/** Handle events */
static Boolean AppHandleEvent( EventPtr eventP) {
	Word formId;
	FormPtr frmP;
	Boolean handled = false;
	switch (eventP->eType) {
	case frmLoadEvent:
		formId = eventP->data.frmLoad.formID;
		frmP = FrmInitForm(formId);
		FrmSetActiveForm(frmP);
		switch (formId) {
		case MainForm:
			FrmSetEventHandler(frmP, MainFormHandleEvent);
			handled = true;
			break;
			/*
		case AboutForm:
			FrmSetEventHandler(frmP, AboutFormHandleEvent);
			handled = true;
			break;
			*/
		default:
			break;
		}
		break;
	default:
		break;
	}
	return handled;
}

/*
static Boolean AboutFormHandleEvent(EventPtr eventP) {
	Boolean handled = false;
	FormPtr frmP;
	switch (eventP->eType)	{
	case frmOpenEvent:
		frmP = FrmGetActiveForm();
		FrmDrawForm ( frmP);
		break;
	case ctlSelectEvent:
		if(eventP->data.ctlSelect.controlID == CloseButton) {
			FrmReturnToForm(MainForm);
			handled = true;
		}
		break;
	default:
		break;
	}
	return handled;
}
*/

/** event handling loop */
static void AppEventLoop(void) {
	Word error;
	EventType event;
	Long wait;
	do {
		wait = (mode == PlaybackMode) ? timeToNewNote - TimGetTicks()
									  : SysTicksPerSecond();
	    if(wait < 0) wait = 0; // if the time has gone, do soon
		EvtGetEvent(&event, wait);
		SysHandleEvent(&event)
		|| MenuHandleEvent(0, &event, &error)
		|| AppHandleEvent(&event)
		|| FrmDispatchEvent(&event);
	} while (event.eType != appStopEvent);
}

/**
 * Get the current application's preferences.
 * @return Err value 0 if nothing went wrong
 */
static Err AppStart(void) {
	Word prefsSize;
	VoidHand handle;
	// Read the saved preferences / saved-state information.
	/*
	prefsSize = sizeof(MyPreferenceType);
	if (PrefGetAppPreferences(appFileCreator, appPrefID, &prefs, &prefsSize, true) !=
			noPreferenceFound) {
	}
	*/
	/*
	KeyRates(false, &KeyRate.initDelay, &KeyRate.period, &KeyRate.doubleTapDelay,
		&KeyRate.queueAhead);
		*/
	// get MaxRecordLength from string resource
	if((handle = DmGetResource('tSTR',MaxRecordLength)) != NULL) {
		CharPtr consPtr = MemHandleLock(handle);
		maxRecordLength = StrAToI(consPtr);
		MemHandleUnlock(handle);
		DmReleaseResource(handle);
	}

	recordHandle = MemHandleNew(sizeof(Note) * (maxRecordLength + 1));
	if(recordHandle == NULL) {
		RECORDABLE = false;
		return 0; // not critical
	}
	record = MemHandleLock(recordHandle);
	RECORDABLE = true;
	recordedLength = 0;
	return 0;
}


/** Save the current state of the application */
static void AppStop(void) {
	SndDoCmd(NULL, &offCmd, -1);
	// Write the saved preferences / saved-state information.  This data
	// will be backed up during a HotSync.
	/*
	PrefSetAppPreferences (appFileCreator, appPrefID, appPrefVersionNum,
		&prefs, sizeof (prefs), true);
		*/
	/*
	KeyRates(true, &KeyRate.initDelay, &KeyRate.period, &KeyRate.doubleTapDelay,
		&KeyRate.queueAhead);
		*/
	FrmCloseAllForms();
	if(RECORDABLE) {
		MemHandleUnlock(recordHandle);
		MemHandleFree(recordHandle);
	}
}

/**
 * This is the main entry point for the application.
 * @param cmd		  Word value specifying the launch code.
 * @param cmdPB		  Pointer to a structure that is associated with the launch code.
 * @param launchFlags Word value providing extra information about the launch.
 * @return Result of launch
 */
DWord PilotMain(Word cmd, Ptr cmdPBP, Word launchFlags) {
	Err error;
	switch (cmd) {
	case sysAppLaunchCmdNormalLaunch:
		error = RomVersionCompatible (ourMinVersion, launchFlags);
		if (error) return error;
		error = AppStart();
		if (error) return error;
		FrmGotoForm(MainForm);
		AppEventLoop();
		AppStop();
		break;
	default:
		break;
	}
	return 0;
}
