/**
 * Sound Pad - keyboard on palm
 * Copyright (C) 2000 by matobaa
 * @version 0.12 2000/12/15
 * @version 0.11 2000/12/10
 * @version 0.1 2000/12/08
 * @author matobaa	mailto:matobaa@mail.dddd.ne.jp
 * @see http://www2u.biglobe.ne.jp/~matobaa/palm/palmware.html
 *
 * sustain concept and UI by P-Key by Koichiro Nakagawa
 * http://hp.vector.co.jp/authors/VA017414/
 */

#include <Pilot.h>
#include <SysEvtMgr.h>
#include <KeyMgr.h>
#include "resource.h"

// From SonyChars.h
#define vchrJogUp		   0x1700
#define vchrJogDown		   0x1701

#define ourMinVersion 3
#define CenterC 60 /* by midi# */
#define mute 3 /* ticks */

#define TONE_KOEF 1.05946309435929530984 /* log2(1/12) */

/* inline functions */
#define included(x,y,w,h,tx,ty) ((x) < (tx) && (tx) < (x+w) && (y) < (ty) && (ty) < (y+h))

/* Internal Structures */
typedef struct {} MyPreferenceType;
typedef struct {} MyAppInfoType;
typedef enum { C = 0, Cis, D, Dis, E, F, Fis, G, Gis, A, Ais, H } Note;
#define Off -1
#define Volume -2

/* Internal Variables */
static SndCommandType onCmd = {sndCmdNoteOn, 60,2200,127};
static SndCommandType offCmd = {sndCmdQuiet, 0,0,0};
static MyPreferenceType prefs;
static MyAppInfoType appInfo;
static UInt offset = CenterC;

/* Internal Constants */
static const unsigned long appFileCreator = 'SndP';
static const unsigned short appVersionNum = 0x01;
static const unsigned short appPrefID = 0x00;
static const unsigned short appPrefVersionNum = 0x01;

/* prototypes */
static Err RomVersionCompatible(DWord , Word );
static VoidPtr GetObjectPtr(Word);
static int getNote(UInt,UInt);
static void DrawKeyboard();

/* functions */

/** Check the rom version */
static Err RomVersionCompatible(DWord requiredVersion, Word launchFlags) {
	DWord romVersion;
	// See if we're on in minimum required version of the ROM or later.
	FtrGet(sysFtrCreator, sysFtrNumROMVersion, &romVersion);
	if (romVersion < requiredVersion) {
		if ((launchFlags & (sysAppLaunchFlagNewGlobals | sysAppLaunchFlagUIApp)) ==
				(sysAppLaunchFlagNewGlobals | sysAppLaunchFlagUIApp)) {
			FrmAlert (SystemIncompatible);
			// Pilot 1.0 will continuously relaunch this app unless we switch to
			// another safe one.
			if (romVersion < sysMakeROMVersion(3,0,0,sysROMStageRelease,0)) {
				AppLaunchWithCommand(sysFileCDefaultApp, sysAppLaunchCmdNormalLaunch, NULL);
			}
		}
		return sysErrRomIncompatible;
	}
	return 0;
}

/** Get the UI Oeject's Ptr specified by ID */
static VoidPtr GetObjectPtr(Word objectID) {
	FormPtr frmP;
	frmP = FrmGetActiveForm();
	return (FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, objectID)));
}

#define KBPosX 3
#define KBPosY 19
#define IvoryWidth 22
#define EbonyWidth 15
#define EbonyHeight 42
#define IvoryHeight 76
#define KBWidth (IvoryWidth * 7)
#define KBHeight IvoryHeight

/** get key ID by tapped location */
static int getNote(UInt x,UInt y) {
	int roundedOffset = offset - offset % 12;
	if (included(68,127,24,24,x,y)) return Volume;
	// else
	x += ((offset % 12) + 1 & ~1) * 22 / 2; // re-align x by offset
	if(x > KBPosX + KBWidth) { x -= KBWidth ; roundedOffset += 12;} // fold by 3 to 156
	if(x < KBPosX) { x += KBWidth ; roundedOffset -= 12;}
	return included( 16,KBPosY,13,EbonyHeight,x,y) ? Cis + roundedOffset :
		   included( 42,KBPosY,13,EbonyHeight,x,y) ? Dis + roundedOffset :
		// included( ??.KBPosY, 0,EbonyHeight,x,y) ? Eis + roundedOffset :
		   included( 81,KBPosY,13,EbonyHeight,x,y) ? Fis + roundedOffset :
		   included(106,KBPosY,13,EbonyHeight,x,y) ? Gis + roundedOffset :
		   included(131,KBPosY,13,EbonyHeight,x,y) ? Ais + roundedOffset :
		   included(  3,KBPosY,21,IvoryHeight,x,y) ? C + roundedOffset :
		   included( 25,KBPosY,21,IvoryHeight,x,y) ? D + roundedOffset :
		   included( 47,KBPosY,21,IvoryHeight,x,y) ? E + roundedOffset :
		   included( 69,KBPosY,21,IvoryHeight,x,y) ? F + roundedOffset :
		   included( 91,KBPosY,21,IvoryHeight,x,y) ? G + roundedOffset :
		   included(113,KBPosY,21,IvoryHeight,x,y) ? A + roundedOffset :
		   included(135,KBPosY,21,IvoryHeight,x,y) ? H + roundedOffset : Off ;
}

/* get Octave shift offset */
static unsigned int TempOffset() {
	int currentState = KeyCurrentState();
	return ((currentState & keyBitHard1) != 0) * -24
		 + ((currentState & keyBitHard2) != 0) * -12
		 + ((currentState & keyBitHard3) != 0) *  12
		 + ((currentState & keyBitHard4) != 0) *  24;
}

/* local inline functions */
#define noteContinue() SndDoCmd(NULL, (currentNote != Off ) ? &onCmd : &offCmd, -1)
#define noteOn() onCmd.param1 = note + TempOffset(); SndDoCmd(NULL, &onCmd, -1); currentNote = note
#define noteOff() SndDoCmd(NULL, &offCmd, -1); currentNote = Off;

/** Handle events at main form */
static Boolean MainFormHandleEvent(EventPtr eventP) {
	static int currentNote = Off;
	FormPtr frmP;
	int note = getNote(eventP->screenX, eventP->screenY);
	Boolean handled = false;
	switch (eventP->eType)	{
	case frmOpenEvent:
		frmP = FrmGetActiveForm();
		FrmDrawForm ( frmP);
		DrawKeyboard();
		break;
	case keyDownEvent:
		if(eventP->data.keyDown.chr == chrPageUp ||
		   eventP->data.keyDown.chr == vchrJogUp ) {
			if(offset <= 127 - 12 - 2) {
				offset += 2;
				if(offset % 12 == 6) offset--;
				if(offset % 12 == 1) offset--;
				DrawKeyboard();
			}
		} else
		if(eventP->data.keyDown.chr == chrPageDown ||
		   eventP->data.keyDown.chr == vchrJogDown ) {
			if(offset >= 2) {
				offset -= 2;
				if(offset % 12 == 3) offset++;
				if(offset % 12 == 10) offset++;
				DrawKeyboard();
			}
		}
//	  |1|3|-|6|8|A|
//	 |0|2|4|5|7|9|B|
		handled = true;
		if (eventP->penDown &&
				(note = getNote(eventP->screenX, eventP->screenY)) > Off ) {
			noteOn(); // now, we have new offset value: recalculate note.
		} else {
			noteContinue();
		}
		break;
	case penDownEvent:
		if (note == Off) { // tapped on outside
			noteOff();
		} else if (note == Volume) {
			// PENDING : reserved position for future
		} else {
			if ( note == currentNote ) { // tapped same key
				// quite mute
				SndDoCmd(NULL, &offCmd, -1);
				SysTaskDelay(mute);
			}
			// play a note
			noteOn();
		}
		handled = true;
		break;
	case penMoveEvent:
		if ( note > Off && note != currentNote) { // move to other key
			noteOn();
		} else {
			noteContinue();
		}
		handled = true;
		break;
	case penUpEvent:
		if ( note == currentNote) { // if release on same key
			noteOff(); // stop note
		} else { // if release on outside
			noteContinue(); // sustain
		}
		handled = true;
		break;
	case nilEvent:
		handled = true;
		noteContinue();
		break;
	default:
		noteContinue();
		break;
	}
	return handled;
}

#undef noteContinue()
#undef noteOn()
#undef noteOff()

/**
 * Draw ebony keys in specified location
 */
static void WinDrawEbony(int x){
	RectangleType rect;
	while(x > KBPosX + KBWidth ) x -= KBWidth;
	rect.topLeft.x = x;  rect.extent.x = EbonyWidth;
	rect.topLeft.y = KBPosY; rect.extent.y = EbonyHeight ;
	if(x > KBPosX + KBWidth - EbonyWidth ) { // if right most key,
		// draw additional key in leftside
		rect.topLeft.x = 0;
		rect.extent.x = EbonyWidth - (KBWidth - x);
		WinDrawRectangle(&rect,1);
		// and restore location for original key
		rect.topLeft.x = x;
		rect.extent.x = (x < 160 - EbonyWidth) ? EbonyWidth : 160 - x;
	}
	WinDrawRectangle(&rect,1);
	WinEraseLine( x+2, 24, x+2, 58);
}

static void DrawKeyboard() {
	int i;
	RectangleType rect;
	char buffer;
	int roundedOffset = ((offset % 12) + 1 & ~1 )/ 2;
	// erase
	rect.topLeft.x = 0; rect.topLeft.y = KBPosY;
	rect.extent.x = 160; rect.extent.y = IvoryHeight;
	WinEraseRectangle(&rect,0);
	// Ebony
	WinDrawEbony( 16 + KBWidth - roundedOffset * 22); // Cis
	WinDrawEbony( 42 + KBWidth - roundedOffset * 22); // Dis
	WinDrawEbony( 81 + KBWidth - roundedOffset * 22); // Fis
	WinDrawEbony(106 + KBWidth - roundedOffset * 22); // Gis
	WinDrawEbony(131 + KBWidth - roundedOffset * 22); // Ais
	// Ivory
	WinDrawLine(0,KBPosY - 1,159,KBPosY - 1);
	WinDrawLine(0,KBPosY + IvoryHeight + 1,159,KBPosY + IvoryHeight + 1);
	for(i=3; i<=157; i+=22) {
		WinDrawLine( i, KBPosY, i, KBPosY + IvoryHeight);
	}
	// C Mark
	rect.topLeft.x = (roundedOffset == 0) ? 12
					: 12 + KBWidth - roundedOffset * IvoryWidth;
	rect.topLeft.y = 85;
	rect.extent.x  = 4; rect.extent.y = 4;
	//WinDrawRectangle(&rect,1);
	buffer = '0' + ((offset-1) / 12); // little tricky
	WinDrawChars(&buffer,1,rect.topLeft.x, rect.topLeft.y);
}

/** Handle events */
static Boolean AppHandleEvent( EventPtr eventP) {
	Word formId;
	FormPtr frmP;
	Boolean handled = false;
	switch (eventP->eType) {
	case frmLoadEvent:
		formId = eventP->data.frmLoad.formID;
		frmP = FrmInitForm(formId);
		FrmSetActiveForm(frmP);
		switch (formId) {
		case MainForm:
			FrmSetEventHandler(frmP, MainFormHandleEvent);
			handled = true;
			break;
			/*
		case AboutForm:
			FrmSetEventHandler(frmP, AboutFormHandleEvent);
			handled = true;
			break;
			*/
		default:
			break;
		}
		break;
	default:
		break;
	}
	return handled;
}

/*
static Boolean AboutFormHandleEvent(EventPtr eventP) {
	Boolean handled = false;
	FormPtr frmP;
	switch (eventP->eType)	{
	case frmOpenEvent:
		frmP = FrmGetActiveForm();
		FrmDrawForm ( frmP);
		break;
	case ctlSelectEvent:
		if(eventP->data.ctlSelect.controlID == CloseButton) {
			FrmReturnToForm(MainForm);
			handled = true;
		}
		break;
	default:
		break;
	}
	return handled;
}
*/

/** event handling loop */
static void AppEventLoop(void) {
	Word error;
	EventType event;
	do {
		EvtGetEvent(&event, SysTicksPerSecond());
		SysHandleEvent(&event)
		|| MenuHandleEvent(0, &event, &error)
		|| AppHandleEvent(&event)
		|| FrmDispatchEvent(&event);
	} while (event.eType != appStopEvent);
}

/**
 * Get the current application's preferences.
 * @return Err value 0 if nothing went wrong
 */
static Err AppStart(void) {
	Word prefsSize;
	// Read the saved preferences / saved-state information.
	/*
	prefsSize = sizeof(MyPreferenceType);
	if (PrefGetAppPreferences(appFileCreator, appPrefID, &prefs, &prefsSize, true) !=
			noPreferenceFound) {
	}
	*/
	KeySetMask(~(keyBitHard1 | keyBitHard2 | keyBitHard3 | keyBitHard4));
	/*
	KeyRates(false, &KeyRate.initDelay, &KeyRate.period, &KeyRate.doubleTapDelay,
		&KeyRate.queueAhead);
		*/
	return 0;
}


/** Save the current state of the application. */
static void AppStop(void) {
	SndDoCmd(NULL, &offCmd, -1);
	// Write the saved preferences / saved-state information.  This data
	// will be backed up during a HotSync.
	/*
	PrefSetAppPreferences (appFileCreator, appPrefID, appPrefVersionNum,
		&prefs, sizeof (prefs), true);
		*/
	KeySetMask( keyBitsAll );
	/*
	KeyRates(true, &KeyRate.initDelay, &KeyRate.period, &KeyRate.doubleTapDelay,
		&KeyRate.queueAhead);
		*/
}

/**
 * This is the main entry point for the application.
 * @param cmd		  Word value specifying the launch code.
 * @param cmdPB		  Pointer to a structure that is associated with the launch code.
 * @param launchFlags Word value providing extra information about the launch.
 * @return Result of launch
 */
DWord PilotMain(Word cmd, Ptr cmdPBP, Word launchFlags) {
	Err error;
	switch (cmd) {
	case sysAppLaunchCmdNormalLaunch:
		error = RomVersionCompatible (ourMinVersion, launchFlags);
		if (error) return error;
		error = AppStart();
		if (error) return error;
		FrmGotoForm(MainForm);
		AppEventLoop();
		AppStop();
		break;
	default:
		break;
	}
	return 0;
}
